from pathlib import Path
from macrostrat.app_frame import Application


APP_ROOT = Path(__file__).parent

app = Application(
    "Test App",
    restart_commands={"gateway": "caddy reload --config /etc/caddy/Caddyfile"},
    app_module="test_app",
    compose_files=[APP_ROOT / "docker-compose.yaml"],
)
main = app.control_command()



import fcntl
import os
import sys
import termios
from contextlib import contextmanager
from enum import Enum
from subprocess import Popen
from time import sleep

from macrostrat.utils import get_logger

from .core import Application

log = get_logger(__name__)


def follow_logs(app: Application, container: str, **kwargs):
    app.info("Following container logs", style="green bold")
    app.info(
        f"- Press [bold]q[/bold] or [bold]Ctrl+c[/bold] to exit logs (:app_name: will keep running)."
    )
    app.info(
        f"- Press [bold]r[/bold] or run [cyan]:command_name: restart[/cyan] to restart :app_name:."
    )
    app.info(
        f"- Press [bold]x[/bold] or run [cyan]:command_name: down[/cyan] to stop :app_name:.",
    )
    # Should integrate this into the macrostrat.utils.cmd function
    # env = kwargs.pop("env", _build_compose_env())
    args = ["docker", "compose", "logs", "-f", "--since=1s"]
    log.debug(" ".join(args))
    sleep(0.1)
    app.console.print()

    if container is not None and container != "":
        args.append(container)
    return Popen(args, **kwargs)


class Result(Enum):
    CONTINUE = 1
    RESTART = 2
    EXIT = 3


def follow_logs_with_reloader(
    app: Application,
    container: str,
) -> Result:
    proc = follow_logs(app, container)
    try:
        with wait_for_keys():
            while True:
                # Read input from stdin
                try:
                    key = sys.stdin.read(1)
                    if key == "q" or key == "Q":
                        return Result.CONTINUE
                    elif key == "r" or key == "R":
                        return Result.RESTART
                    elif key == "x" or key == "X":
                        return Result.EXIT
                except IOError:
                    pass
    finally:
        proc.kill()
        proc.wait()


@contextmanager
def wait_for_keys():
    fd = sys.stdin.fileno()

    oldterm = termios.tcgetattr(fd)
    newattr = termios.tcgetattr(fd)
    newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
    termios.tcsetattr(fd, termios.TCSANOW, newattr)

    oldflags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, oldflags | os.O_NONBLOCK)

    try:
        yield
    finally:
        termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
        fcntl.fcntl(fd, fcntl.F_SETFL, oldflags)



# Typer command-line application

import sys
from time import sleep

import click
import typer
from click import Group
from macrostrat.utils import get_logger
from typer import Context, Typer
from typer.core import TyperGroup
from typer.models import TyperInfo

from .compose import check_status, compose
from .core import Application
from .follow_logs import Result, follow_logs_with_reloader

log = get_logger(__name__)


class OrderCommands(TyperGroup):
    def list_commands(self, ctx: Context):
        """Return list of commands in the order of appearance."""
        return list(self.commands)  # get commands using self.commands


class ControlCommand(Typer):
    name: str

    app: Application
    _click: Group

    def __init__(
        self,
        app: Application,
        **kwargs,
    ):
        kwargs.setdefault("add_completion", False)
        kwargs.setdefault("no_args_is_help", True)
        kwargs.setdefault("cls", OrderCommands)
        kwargs.setdefault("name", app.name)
        super().__init__(**kwargs)
        self.app = app
        self.name = app.name

        def callback(ctx: Context, verbose: bool = False):
            ctx.obj = self.app
            self.app.setup_logs(verbose=verbose)

        callback.__doc__ = f"""{self.app.name} command-line interface"""

        self.registered_callback = TyperInfo(callback=callback)

        # Click commands must be added after Typer commands in the current design.
        self._click_commands = []

        self.build_commands()

    def build_commands(self):
        for cmd in [up, down, restart]:
            if cmd.__doc__ is not None:
                cmd.__doc__ = self.app.replace_names(cmd.__doc__)
            self.command(rich_help_panel="System")(cmd)
        self.add_click_command(_compose, "compose", rich_help_panel="System")

    def add_click_command(self, cmd, *args, **kwargs):
        """Add a click command for lazy initialization
        params:
            cmd: click command
            args: args to pass to click.add_command
            kwargs: kwargs to pass to click.add_command
            rich_help_panel: name of rich help panel to add to
        """
        rich_help_panel = kwargs.pop("rich_help_panel", None)
        if rich_help_panel is not None:
            setattr(cmd, "rich_help_panel", rich_help_panel)
        cfunc = lambda _click: _click.add_command(cmd, *args, **kwargs)
        self._click_commands.append(cfunc)

    def __call__(self):
        """Run this command using its underlying click object."""
        cmd = typer.main.get_command(self)
        assert isinstance(cmd, click.Group)
        self._click = cmd
        for cfunc in self._click_commands:
            cfunc(self._click)
        return self._click()


def up(
    ctx: Context, container: str = typer.Argument(None), force_recreate: bool = False
):
    """Start the :app_name: server and follow logs."""
    app = ctx.find_object(Application)
    if app is None:
        raise ValueError("Could not find application config")
    if container is None:
        _container = ""

    compose("build", _container)

    sleep(0.1)

    res = compose(
        "up",
        "--no-start",
        "--remove-orphans",
        "--force-recreate" if force_recreate else "",
        _container,
    )
    if res.returncode != 0:
        app.info(
            "One or more containers did not build successfully, aborting.",
            style="red bold",
        )
        sys.exit(res.returncode)
    else:
        app.info("All containers built successfully.", style="green bold")
    print()

    # Get list of currently running containers
    running_containers = check_status(app.name, app.command_name)

    app.info("Starting :app_name: server...", style="bold")
    compose("start")
    print()

    run_restart_commands(app, running_containers)

    res = follow_logs_with_reloader(app, _container)
    if res == Result.RESTART:
        app.info("Restarting :app_name: server...", style="bold")
        ctx.invoke(up, ctx, container)
    elif res == Result.EXIT:
        app.info("Stopping :app_name: server...", style="bold")
        ctx.invoke(down, ctx)
    elif res == Result.CONTINUE:
        app.info(
            "[bold]Detaching from logs[/bold] [dim](:app_name: will continue to run)[/dim]",
            style="bold",
        )
        return


def run_restart_commands(app, running_containers):
    for c, command in app.restart_commands.items():
        if c in running_containers:
            app.info(f"Reloading {c}...", style="bold")
            compose("exec", c, command)
    print()


def down(ctx: Context):
    """Stop all :app_name: services."""
    app = ctx.find_object(Application)
    if app is None:
        raise ValueError("Could not find application config")
    app.info("Stopping :app_name: server...", style="bold")
    compose("down", "--remove-orphans")


def restart(ctx: Context, container: str = typer.Argument(None)):
    """Restart the :app_name: server and follow logs."""
    ctx.invoke(up, ctx, container, force_recreate=True)


@click.command(
    "compose",
    context_settings=dict(
        ignore_unknown_options=True,
        help_option_names=[],
        max_content_width=160,
        # Doesn't appear to have landed in Click 7? Or some other reason we can't access...
        # short_help_width=160,
    ),
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def _compose(args):
    """Run docker compose commands in the appropriate context"""
    compose(*args)



from rich.console import Console
from typing import Optional, Callable
from macrostrat.utils import setup_stderr_logs, get_logger
import logging
from os import environ
from pathlib import Path
from dotenv import load_dotenv


log = get_logger(__name__)


class ApplicationBase:
    name: str
    command_name: str
    app_module: Optional[str]
    root_dir: Path


EnvironmentDependency = dict[str, str] | Callable[[ApplicationBase], dict[str, str]]
ComposeFilesDependency = list[Path] | Callable[[ApplicationBase], list[Path]]


class Application(ApplicationBase):
    console: Console
    _dotenv_cfg: bool | Path | list[Path]

    def __init__(
        self,
        name: str,
        *,
        command_name: Optional[str] = None,
        project_prefix: Optional[str] = None,
        restart_commands: dict[str, str] = {},
        app_module: Optional[str] = None,
        root_dir: Path | Callable[[Path], Path] = Path.cwd(),
        compose_files: ComposeFilesDependency = [],
        env: EnvironmentDependency = {},
        load_dotenv: bool | Path | list[Path] = False,
    ):
        self.name = name
        self.command_name = command_name or name.lower()
        self.project_prefix = project_prefix or name.lower().replace(" ", "_")
        self.console = Console()
        self.restart_commands = restart_commands
        self.app_module = app_module
        self._dotenv_cfg = load_dotenv

        # Root dir and compose files can be specified using dependency injection.
        if callable(root_dir):
            root_dir = root_dir(Path.cwd())
        self.root_dir = root_dir

        if callable(compose_files):
            compose_files = compose_files(self)
        self.compose_files = compose_files
        # Environment setup should possibly be postponed until within a command context.
        self.setup_environment(env)

    def replace_names(self, text: str) -> str:
        text = text.replace(":app_name:", self.name)
        return text.replace(":command_name:", self.name.lower())

    def info(self, text, style=None):
        self.console.print(self.replace_names(text), style=style)

    def load_dotenv(self):
        if isinstance(self._dotenv_cfg, list):
            for path in self._dotenv_cfg:
                load_dotenv(path)
        elif isinstance(self._dotenv_cfg, Path):
            load_dotenv(self._dotenv_cfg)
        elif load_dotenv is True:
            load_dotenv()

    def setup_environment(self, env: EnvironmentDependency):
        environ["DOCKER_SCAN_SUGGEST"] = "false"
        environ["DOCKER_BUILDKIT"] = "1"

        # Set up environment for docker-compose
        # We may need to move this to a context where it is only
        # applied for docker-compose commands
        environ["COMPOSE_PROJECT_NAME"] = self.project_prefix
        compose_files = ":".join([str(f) for f in self.compose_files])
        environ["COMPOSE_FILE"] = compose_files

        # Additional user-specified environment variables
        if callable(env):
            env = env(self)
        for k, v in env.items():
            environ[k] = v

    def setup_logs(self, verbose: bool = False):
        if self.app_module is None:
            log.warning("No app module specified, not setting up logs")
            return
        if verbose:
            setup_stderr_logs(self.app_module)
        else:
            # Disable all logging
            # TODO: This is a hack, we shouldn't have to explicitly disable
            # logging in the CLI. Perhaps there's somewhere that it's being
            # enabled that we haven't chased down?
            setup_stderr_logs("", level=logging.CRITICAL)

    def control_command(self, *args, **kwargs):
        from .control_command import ControlCommand

        return ControlCommand(self, *args, **kwargs)



from macrostrat.utils import cmd, get_logger
from rich.console import Console

console = Console()

log = get_logger(__name__)


def compose(*args, **kwargs):
    """Run docker compose commands in the appropriate context"""
    return cmd("docker", "compose", *args, **kwargs)


def check_status(app_name: str, command_name: str) -> list[str]:
    # Check if containers are running
    res = compose("ps --services --filter status=running", capture_output=True)
    running_containers = res.stdout.decode("utf-8").strip()
    if running_containers == "":
        return []

    containers = running_containers.split("\n")
    containers = [c.strip() for c in containers]

    console.print("[dim]Some containers are already running and up to date: ")
    console.print("  " + ", ".join(containers))
    console.print(
        f"[dim]To fully restart {app_name}, run [cyan]{command_name} restart[/cyan]"
        f" or [cyan]{command_name} up --force-recreate[/cyan]."
    )

    return containers
